"use strict";(self.webpackChunknbgrader=self.webpackChunknbgrader||[]).push([[376,236],{376:(e,r,t)=>{t.r(r),t.d(r,{INotebookTree:()=>s,NotebookTreeWidget:()=>n});var a=t(888),o=t(778);class n extends o.TabPanel{constructor(){super({tabPlacement:"top",tabsMovable:!0,renderer:a.TabBarSvg.defaultRenderer}),this.addClass("jp-TreePanel")}}const s=new(t(930).Token)("@jupyter-notebook/tree:INotebookTree")}}]);